from typing import Dict, Any, Union, List
import warnings


# Generate using ChatGPT https://chat.openai.com/
def filter_dict(d: Dict[str, Any], keys_to_exclude: Union[str, List[str]], required_key: str = None) -> Union[Dict[str, Any], None]:
    """
    Returns a new dictionary that contains all the key-value pairs from `d`, except for those whose keys are in `keys_to_exclude`.

    If `required_key` is provided and it exists in `d`, the function will return a new dictionary that contains all the key-value pairs from `d`, except for those whose keys are in `keys_to_exclude`. If `required_key` is not provided or it does not exist in `d`, the function will return None.

    Args:
    - d: A dictionary whose keys are strings and whose values can be of any type.
    - keys_to_exclude: A string or list of strings representing the keys to be excluded from the output dictionary.
    - required_key: An optional string representing a key that must exist in `d` for the function to return a dictionary. If `required_key` is not provided or it does not exist in `d`, the function will return None.

    Returns:
    - A new dictionary that contains all the key-value pairs from `d`, except for those whose keys are in `keys_to_exclude`, or None if `required_key` is not provided or it does not exist in `d`.
    """
    if isinstance(keys_to_exclude, str):
        keys_to_exclude = [keys_to_exclude]
    if required_key is not None and required_key not in d:
        return None
    return {k: v for k, v in d.items() if k not in keys_to_exclude}


class OrkgResponse(object):

    @property
    def succeeded(self) -> bool:
        return str(self.status_code)[0] == '2'

    def __init__(self, response, status_code: str, content: Union[list, dict], url: str, paged: bool):
        if response is None and status_code is None and content is None and url is None:
            raise ValueError("either response should be provided or content with status code")
        if not paged:
            warnings.warn("You are running an out-of-date ORKG backend! GET (i.e., listing) calls now provide pagination information", DeprecationWarning)
        if (response is not None and str(response.status_code)[0] != '2') or (status_code is not None and content is not None and status_code[0] != '2'):
            self.status_code = response.status_code if response is not None else status_code
            self.content = response.content if response is not None else content
            self.page_info = None
            return
        if response is not None:
            self.status_code = response.status_code
            self.content = response.json() if len(response.content) > 0 else response.content
            self.page_info = None
            if paged and isinstance(self.content, dict):
                self.page_info = filter_dict(self.content, 'content', 'pageable')
                self.content = self.content['content'] if 'content' in self.content else self.content
            self.url = response.url
        if status_code is not None and content is not None:
            self.status_code = status_code
            self.content = content
            self.page_info = None
            if paged and isinstance(self.content, dict) and 'content' in content:
                self.page_info = filter_dict(self.content, 'content', 'pageable')
                self.content = self.content['content']
            self.url = url

    def __repr__(self):
        return "%s %s" % ("(Success)" if self.succeeded else "(Fail)", self.content)

    def __str__(self):
        return self.content


class OrkgUnpaginatedResponse(object):

    @property
    def all_succeeded(self) -> bool:
        return all([response.succeeded for response in self.responses])

    def __init__(self, responses: List[OrkgResponse]):
        self.responses = responses
        self.content: List[dict] = []

        for response in self.responses:
            self.content.extend(response.content)



from typing import Optional
import re


class OID(object):
    """
    Represents an ORKG ID
    """

    id: str
    type: Optional[str]

    def __init__(self, id: str, type: Optional[str] = None):
        """
        :param id: the id
        :param type: the type of the id (optional)
        """
        self.id = id
        self.type = type

    def __str__(self):
        return self.id

    def __repr__(self):
        return self.id

    def __eq__(self, other):
        return self.id == other.id and self.type == other.type

    def __hash__(self):
        return hash(self.id)

    def __bool__(self):
        return bool(self.id)

    def is_literal_id(self) -> bool:
        return re.match(r'^L\d+$', self.id) is not None

    def is_resource_id(self) -> bool:
        return re.match(r'^R\d+$', self.id) is not None

    def is_predicate_id(self) -> bool:
        return re.match(r'^P\d+$', self.id) is not None

    def is_class_id(self) -> bool:
        return re.match(r'^C\d+$', self.id) is not None

    def is_custom_id(self) -> bool:
        return not self.is_literal_id() and not self.is_resource_id() and not self.is_predicate_id() and not self.is_class_id()

    def to_orkg_json(self) -> dict:
        obj = {'@id': self.id}
        if self.type:
            obj['@type'] = self.type
        return obj



from typing import Union

from orkg.common import OID
from orkg.out import OrkgResponse
from orkg.utils import NamespacedClient
from orkg.client.harvesters import doi as doi_harvester


class HarvestersClient(NamespacedClient):

    def doi_harvest(self, doi: str, orkg_rf: Union[str, OID]) -> OrkgResponse:
        """
        Harvests DOI data for a paper and add it to the ORKG.
        It works under the assumption that the paper contains some JSON-LD representation of its content
        :param doi: The DOI of the paper to harvest
        :param orkg_rf: The resource ID of the ORKG research field to add the harvested data to, or the string representation to be looked up (can raise errors)
        """
        return doi_harvester.harvest(self.client, doi, orkg_rf)



from orkg.client.templates.codegen import OTFFunctionWriter
from orkg.client.templates.components import Template
from orkg.utils import NamespacedClient
from typing import Dict, List, Optional
from tqdm import tqdm
import json


class TemplatesClient(NamespacedClient):
    materialized_templates: List[str] = []

    def materialize_template(self, template_id: str):
        """
        Materialize a singular ORKG template as python function
        :param template_id: the template id
        :return: True if everything is OK
        """
        materialization_result = self._fetch_and_build_template_function(template_id)
        Template.wip_templates.clear()
        return materialization_result

    def materialize_templates(self, templates: Optional[List[str]] = None, verbose: bool = True):
        """
        Materialize a list of templates (or all templates if not provided)
        :param templates: optional list of templates
        :param verbose: sets if the process shows a progress bar or suppresses it
        """
        if templates is None:
            templates = [template['id'] for template in
                         self.client.classes.get_resource_by_class(class_id='NodeShape', size=1000).content]
        exclude = ['R12000', 'R38504', 'R108555', 'R111221', 'R111231', 'R70785', 'R48000', 'R111155']  # RF, RP, Comp last 3
        iterator = tqdm(templates, desc='Materializing templates') if verbose else templates
        for template_id in iterator:
            if template_id not in exclude:
                self.materialize_template(template_id)

    def _fetch_and_build_template_function(self, template_id: str):
        """
        Internal function to create python digital twins of ORKG templates
        :param template_id: template ID to build
        :return: True if everything is OK
        """
        OTFFunctionWriter.implement_function(self, template_id)
        return True

    def list_templates(self):
        """
        List the set of materialized template functions
        :return: list of strings
        """
        return self.materialized_templates

    def get_template_specifications(self, template_id: str) -> str:
        """
        Return JSON representation of a template's specification (schema)
        :param template_id: the template to lookup
        :return: string representation of a JSON object
        """
        template = Template(self.client, template_id)
        result = {comp.get_clean_name(): f'A value of type ({comp.value_class_label})' for comp in template.components}
        return json.dumps(result, indent=4, sort_keys=True, default=str)

    def create_template_instance(self, template_id: str, instance: Dict) -> str:
        """
        Creates an instance of a given template by filling in
        the specifications of the template with the provided values

        :param template_id: the string representing the template ID
        :param instance: the json object that contains the components of the template instance
        """
        template = Template(self.client, template_id)
        obj = template.create_api_object()
        for key, value in instance.items():
            obj = obj.replace(key, f'"{value}"')
        json_object = json.loads(obj)
        return self.client.objects.add(params=json_object).content['id']



from orkg.utils import NamespacedClient, simcomp_available
from orkg.out import OrkgResponse
from typing import List, Tuple
import pandas as pd
from ast import literal_eval
from urllib.parse import urlparse, parse_qs
from typing import Union, Tuple
from collections import defaultdict


class ContributionsClient(NamespacedClient):

    @simcomp_available
    def similar(self, cont_id: str) -> OrkgResponse:
        self.client.simcomp._append_slash = True
        response = self.client.simcomp.similar(cont_id).GET()
        return self.client.wrap_response(response=response)

    @simcomp_available
    def compare(self, contributions: List[str], response_hash: str = None) -> OrkgResponse:
        self.client.simcomp._append_slash = False
        params = f'?contributions={",".join(contributions)}'
        if response_hash is not None:
            params = f'{params}&response_hash={response_hash}'
        response = self.client.simcomp.compare.GET(params)
        return self.client.wrap_response(response=response)

    def __get_contributions_from_comparison(self, comparison_id: str) -> Tuple[List[str], str, List[str]]:
        resource = self.client.resources.by_id(comparison_id).content
        if 'Comparison' not in resource['classes']:
            raise ValueError("This id is not for a comparison")
        response = self.client.json.get_json(resource_id=comparison_id).content
        query_params = parse_qs(urlparse(response['data']['url']).query)
        return query_params['contributions'][0], query_params['response_hash'][0], query_params['properties'][0].split(',') if 'properties' in query_params else []

    @simcomp_available
    def compare_dataframe(self, contributions: List[str] = None, comparison_id: str = None, like_ui=True,
                          include_meta=False) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
        ui_props = []
        if contributions is None and comparison_id is None:
            raise ValueError("either provide the contributions, or the comparison ID")
        response_hash = None
        if comparison_id is not None:
            contributions, response_hash, ui_props = self.__get_contributions_from_comparison(comparison_id)
        response = self.compare(contributions=contributions, response_hash=response_hash)
        if not response.succeeded:
            return pd.DataFrame()
        content = response.content
        contributions_list = content['contributions']
        contribution_ids_and_titles = {
            contribution['id']: f"{contribution['title']}/{contribution['contributionLabel']} ({contribution['id']})"
            for contribution in contributions_list
        }
        columns = [contribution_ids_and_titles[contribution['id']] for
                   contribution in contributions_list]
        properties_list = content['properties']
        property_lookup = {prop['id']: prop['label'] for prop in properties_list}
        # create table view of the data
        data = content['data']
        indices = []
        rows = []
        for prop_id, values in data.items():
            indices.append(property_lookup[prop_id])
            row = []
            for index, value in enumerate(values):
                if not value[0]:
                    row.append('')
                else:
                    if len(value) == 1:
                        try:
                            # Try parse the single value into datatype
                            row.append(literal_eval(value[0]['label']))
                        except (SyntaxError, ValueError):
                            # If it fails, that means it is a string
                            row.append(value[0]['label'])
                    else:
                        cell = []
                        # Rather than parsing it in a list comprehension that might fail
                        # Then the whole list will be typed as strings, we parse it one item at a time
                        for v in value:
                            try:
                                # Try parse each value by itself
                                cell.append(literal_eval(v['label']))
                            except (SyntaxError, ValueError):
                                # If it fails, that means it is a string
                                cell.append(v['label'])
                        row.append(cell)
            rows.append(row)
        # create data frame from pieces
        df = pd.DataFrame(rows, columns=columns, index=indices)
        if like_ui and len(ui_props) > 0:
            # remove props that are not displayed
            df = df.drop([p_lbl for p_id, p_lbl in property_lookup.items() if p_id not in ui_props])
            # order df rows by ui props order
            df = df.reindex([property_lookup[p] for p in ui_props if p in property_lookup])
        if include_meta:
            # author, doi, publication month, publication year, url, research field, venue
            meta_property_ids = ['P27', 'P26', 'P28', 'P29', 'url', 'P30', 'HAS_VENUE']
            paper_set = set([f"{contribution['paperId']}" for contribution in contributions_list])
            comparison_meta_data = {}
            # create table view of the data
            for paper in paper_set:
                paper_meta_dict_of_lists = defaultdict(list)
                paper_statements = self.client.statements.get_by_subject_unpaginated(subject_id=paper)
                for statement in paper_statements.content:
                    if statement['predicate']['id'] in meta_property_ids:
                        pred = statement['predicate']['label']
                        obj = statement['object']['label']
                        paper_meta_dict_of_lists[pred].append(obj)
                # make dict values strings if only one and list of strings if two or more
                paper_meta_data = {k: v.pop() if len(v) == 1 else v for k, v in paper_meta_dict_of_lists.items()}
                paper_meta_data['title'] = statement['subject']['label']
                paper_meta_data['paper id'] = paper
                # get IDs of all contributions from this paper which are included in the comparison
                contribution_ids = [str(contribution['id']) for contribution in contributions_list if
                                    contribution['paperId'] == paper]
                # add paper meta info for each contribution to comparison meta info
                for contribution_id in contribution_ids:
                    column_name = contribution_ids_and_titles[contribution_id]
                    paper_dict = paper_meta_data.copy()
                    comparison_meta_data[column_name] = paper_dict
                    comparison_meta_data[column_name]['contribution id'] = contribution_id
            # Make dataframe with same column order as comparison df and replace missing fields with empty string
            df_meta = pd.DataFrame.from_dict(comparison_meta_data)[columns].fillna('')
            return df, df_meta
        return df



from typing import Union, Dict, List, Any

from orkg.common import OID
from orkg.out import OrkgResponse

import re
import uuid
import requests


def _validate_doi(doi: str) -> bool:
    """Check if a string is a valid DOI or a complete DOI URL."""
    if doi is None:
        return False

    # DOI pattern.
    doi_pattern = re.compile(r"^10.\d{4,9}/[-._;()/:A-Z0-9]+$", re.I)

    # DOI URL pattern.
    url_pattern = re.compile(r"^https?://doi\.org/10.\d{4,9}/[-._;()/:A-Z0-9]+$", re.I)

    return bool(doi_pattern.match(doi) or url_pattern.match(doi))


def _process_contribution(
        contribution_json: Union[Dict, List, str],
        resulting_object: Union[Dict, List],
        global_ids: Dict,
        context: Dict
):
    if isinstance(contribution_json, List):
        for item in contribution_json:
            _process_contribution(item, resulting_object, global_ids, context)
    elif isinstance(contribution_json, Dict):
        if isinstance(resulting_object, List):
            current_object = {}
            resulting_object.append(current_object)
            resulting_object = current_object
        for key, value in contribution_json.items():
            if key == "@id":
                global_ids[value] = "_" + str(uuid.uuid4())
                resulting_object["@temp"] = global_ids[value]
            elif key == "@type":
                resulting_object["classes"] = [
                    clazz.rsplit("/", 1)[-1]
                    for clazz in
                    contribution_json[key]
                ]
            elif key == "label":
                resulting_object["label"] = str(contribution_json[key])
            elif key == "@context":
                continue
            else:
                # other keys
                if "values" not in resulting_object:
                    resulting_object["values"] = {}
                if isinstance(value, List) and all(isinstance(item, str) for item in value):
                    resulting_object["values"][key] = [{"text": str(item)} for item in value]
                elif isinstance(value, List) or isinstance(value, Dict):
                    orkg_predicate_id = context[key].rsplit("/", 1)[-1]
                    predicate_values = []
                    resulting_object["values"][orkg_predicate_id] = predicate_values
                    _process_contribution(value, predicate_values, global_ids, context)
                else:
                    # Check if the value starts with "_:"
                    # FIXME: Workaround because the descriptions aren't up to date!!
                    if key not in context:
                        orkg_predicate_id = f"CSVW_{key.title()}"
                    else:
                        orkg_predicate_id = context[key].rsplit("/", 1)[-1]
                    resulting_object["values"][orkg_predicate_id] = []
                    if isinstance(value, str) and value.startswith("_:"):
                        resulting_object["values"][orkg_predicate_id].append({"@id": global_ids[value]})
                    else:
                        resulting_object["values"][orkg_predicate_id].append({"text": str(value)})


def harvest(orkg_client: Any, doi: str, orkg_rf: Union[str, OID]) -> OrkgResponse:
    # TODO: activate after stable DOIs are used
    # # Check if the doi is a valid DOI string
    # if not _validate_doi(doi):
    #     raise ValueError(f'{doi} is not a valid DOI string')

    # Get the content behind the DOI
    url = doi if doi.startswith('http') else f'https://doi.org/{doi}'
    response = requests.get(url, headers={'Accept': 'application/json'})
    if response.status_code != 200:
        raise ValueError(f'Unable to retrieve the content behind the DOI {doi}')
    response = response.json()

    if isinstance(orkg_rf, str):
        rf_response = orkg_client.resources.get(q=orkg_rf, exact=True, size=1)
        if not rf_response.succeeded or len(rf_response.content) == 0:
            raise ValueError(f'Unable to find the ORKG research field with the given string value {orkg_rf}')
        orkg_rf = OID(rf_response.content[0]['id'])

    doi_content = response["data"]
    paper_json = {
        # TODO: handle multiple titles
        "title": doi_content["attributes"]["titles"][0]["title"],
        "doi": doi_content["attributes"]["doi"],
        "authors": [
            {"label": f'{creator["givenName"]} {creator["familyName"]}'}
            for creator in doi_content["attributes"]["creators"]
        ],
        "publicationYear": int(doi_content["attributes"].get("publicationYear")) if doi_content["attributes"].get("publicationYear") else None,
        "publishedIn": doi_content["attributes"].get("publisher", None),
        "researchField": f"{orkg_rf}",
    }
    # get contribution info
    contributions_urls = [
        url["relatedIdentifier"]
        for url in filter(
            lambda x: x["relationType"] == "IsSupplementedBy" and x["relatedIdentifierType"] == "URL",
            doi_content["attributes"]["relatedIdentifiers"]
        )
    ]
    if len(contributions_urls) > 0:
        if "contributions" not in paper_json:
            paper_json["contributions"] = []
        # process each file separately into a contribution
        for url in contributions_urls:
            contribution_json = requests.get(url).json()
            orkg_contribution_json = {}
            global_ids = {}
            context = contribution_json["@context"]
            _process_contribution(contribution_json, orkg_contribution_json, global_ids, context)
            # replace the key "label" with "name"
            orkg_contribution_json["name"] = orkg_contribution_json.pop("label")
            paper_json["contributions"].append(orkg_contribution_json)
    # Now that we have everything, let's finalize the paper object and add it to the graph
    paper_json = {"paper": paper_json}
    return orkg_client.papers.add(paper_json)



import datetime
import keyword
import re
from typing import Union, List

Number = Union[int, float, complex, List[int], List[float], List[complex]]
Boolean = Union[bool, List[bool]]
String = Union[str, List[str]]
Integer = Union[int, List[int]]
Date = Union[datetime.date, List[datetime.date]]


def clean_up_dict(dictionary: dict) -> dict:
    """
    replaces the key 'name' with 'label' inside the 'resource' object.
    then strips the root key 'resource' and returns the inner dictionary.
    """
    dictionary = dictionary['resource']
    dictionary['label'] = dictionary['name']
    del dictionary['name']
    return dictionary


def pre_process_string(string: str) -> str:
    lower_string = string.lower().replace(' ', '_')
    res = re.sub(r'\W+', '', lower_string)
    return res


def check_against_keywords(string: str) -> str:
    """
    Checks a string against python keywords.
    If a keyword is found, it is postfixed with a '_template'.
    """
    if string in keyword.kwlist:
        return string + '_template'
    return string


def map_type_to_pythonic_type(value_type: str) -> str:
    if value_type == 'Number':
        return 'Number'
    elif value_type == 'Boolean':
        return 'Boolean'
    elif value_type == 'Integer':
        return 'Integer'
    elif value_type == 'Date':
        return 'Date'
    else:
        return 'String'


def remove_empty_values(dictionary: dict):
    if 'values' not in dictionary['resource']:
        return
    for _, value in dictionary['resource']['values'].items():
        for v in value:
            if not bool(v):
                value.remove(v)



from orkg.out import OrkgResponse, OrkgUnpaginatedResponse
from orkg.utils import NamespacedClient, query_params, dict_to_url_params


class ContributionComparisonsClient(NamespacedClient):

    @query_params("page", "size", "sort", "desc")
    def by_ids(self, ids, params=None) -> OrkgResponse:
        """
        Get contributions details given by their IDs.

        :param ids: the IDs of the contributions.
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: an OrkgResponse object containing the resource
        """

        self.client.backend._append_slash = False
        params["ids"] = ','.join(ids)
        response = self.client.backend("contribution-comparisons").contributions.GET(dict_to_url_params(params))
        return self.client.wrap_response(response)

    @query_params("page", "size", "sort", "desc")
    def by_ids_unpaginated(self, ids, params=None, start_page=0, end_page=-1) -> OrkgUnpaginatedResponse:
        """
        Get all contributions details by all pages between start_page and end_page given by their IDs.

        :param ids: the IDs of the contributions.
        :param page: the page number (optional)
        :param size: number of items per page (optional)
        :param sort: key to sort on (optional)
        :param desc: true/false to sort desc (optional)
        :return: an OrkgUnpaginatedResponse object
        """
        return self._call_pageable(
            self.by_ids,
            args={
                "ids": ids
            },
            params=params,
            start_page=start_page,
            end_page=end_page
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['orkg',
 'orkg.client',
 'orkg.client.harvesters',
 'orkg.client.templates',
 'orkg.graph']

package_data = \
{'': ['*']}

install_requires = \
['Inflector>=3.1.0,<4.0.0',
 'cardinality>=0.1.1,<0.2.0',
 'hammock>=0.2.4,<0.3.0',
 'networkx>=3.1,<4.0',
 'pandas>=2.0.1,<3.0.0',
 'requests>=2.31.0,<3.0.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'orkg',
    'version': '0.16.1',
    'description': 'The official python client for the Open Research Knowledge Graph (ORKG) API',
    'long_description': '# orkg-pypi\n[![Documentation Status](https://readthedocs.org/projects/orkg/badge/?version=latest)](https://orkg.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/orkg.svg)](https://badge.fury.io/py/orkg)\n\nA python client interacting with the ORKG API and sprinkling some python magic on top.\n\nThe package a implements many of the API calls described in the [documentation](http://tibhannover.gitlab.io/orkg/orkg-backend/api-doc/), and provides a set of extra features like graph pythonic objects and dynamic instantiation of entities from specifications.\n\nYou can find details about how-to use the package on [Read the Docs](https://orkg.readthedocs.io/en/latest/index.html).\n\nSpecial thanks to **Allard Oelen** , **Kheir Eddine Farfar**, and **Omar Arab Oghli** for their contributions to the code.',
    'author': 'Yaser Jaradeh',
    'author_email': 'yaser.jaradeh@tib.eu',
    'maintainer': 'Yaser Jaradeh',
    'maintainer_email': None,
    'url': 'http://orkg.org/about',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
